/**********************************************************************/
/* The Minstrel's Song is (c) 1998 by Jonathan Hayward.  See the file */
/* lib/copyrigh.txt for a copyright and license statement.            */
/**********************************************************************/

#include "tms.h"
#include <math.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

void add_new_light_source_to_creature(struct small_creature *c, long radius)
    {
    struct light_source_struct *light_source;
    light_source = (struct light_source_struct *) calloc(1, sizeof(struct
      light_source_struct));
#ifdef DEBUG
    light_source->me = light_source;
#endif
    light_source->next = current_level->light_sources;
    light_source->holder = c;
    light_source->radius = radius;
    light_source->x = c->x;
    light_source->y = c->y;
    light_source->is_on = TRUE;
    current_level->light_sources = light_source;
    if (radius > 0)
	flag_reset(&c->flags, HIDDEN);
    }

/* Works for one light source.  Returns whether it found one.  */

long remove_creatures_light_source(struct level *l, struct small_creature *c)
    {
    struct light_source_struct *light_source, *traverse, *prev;
    if (l->light_sources != NULL)
	{
	if (l->light_sources->holder == c)
	    {
	    light_source = l->light_sources;
	    l->light_sources = l->light_sources->next;
	    free(light_source);
	    return TRUE;
	    }
	else
	    for(prev = traverse = l->light_sources; traverse != NULL;)
		{
		if (traverse->holder == c)
		    {
		    prev->next = traverse->next;
		    free(traverse);
		    traverse = NULL;
		    return TRUE;
		    }
		prev = traverse;
		if (traverse != NULL)
		    traverse = traverse->next;
		}
	}
    return FALSE;
    }

/* Return the radius of the light source if c has a light source, 0 otherwise.
*/

long find_light_source_radius(struct small_creature *c)
    {
    struct ilist *traverse;
    if (c != NULL)
	{
	if (c->extended != NULL)
	    {
	    if (c->extended->wear_wield != NULL)
		{
		for(traverse = c->extended->wear_wield; traverse != NULL;
		  traverse = traverse->next)
		    if ((orig_items[traverse->current->identity].effect ==
		      GIVES_LIGHT) && (flag_check(traverse->current->flags
		      , ACTIVATED)))
			return orig_items[traverse->current->identity].range;
		}
	    if (flag_check(master_creatures[c->identity].flags,
	      HAS_LIGHT_SOURCE))
		return master_creatures[c->identity].radius;
	    else
		return 0;
	    }
	else if (flag_check(master_creatures[c->identity].flags,
	  HAS_LIGHT_SOURCE))
	    return master_creatures[c->identity].radius;
	else
	    return 0;
	}
    else
	return 0;
    }

struct small_item *find_light_source(struct small_creature *c)
    {
    struct ilist *traverse;
    if (c != NULL)
	{
	if (c->extended != NULL)
	    {
	    if (c->extended->wear_wield != NULL)
		{
		for(traverse = c->extended->wear_wield; traverse != NULL;
		  traverse = traverse->next)
		    if ((orig_items[traverse->current->identity].effect ==
		      GIVES_LIGHT) && (flag_check(orig_items
		      [traverse->current->identity].flags, ACTIVATED)))
			return traverse->current;
		}
	    }
	}
    return NULL;
    }

struct small_creature *creature_has_light_source(struct level *l)
    {
    struct light_source_struct *traverse;
    if (l->light_sources != NULL)
	for(traverse = l->light_sources; traverse != NULL; traverse =
	  traverse->next)
	    if (traverse->holder != NULL)
		return traverse->holder;
    return NULL;
    }

void remove_creature_light_sources(struct level *l)
    {
    while (creature_has_light_source(l))
	remove_creatures_light_source(l, creature_has_light_source(l));
    }

void recalculate_light_sources(struct level *l)
    {
    struct clist *traverse;
    remove_creature_light_sources(l);
    if (l->creatures != NULL)
	for(traverse = l->creatures; traverse != NULL; traverse =
	  traverse->next)
	    if (find_light_source_radius(traverse->current) > 0)
		add_new_light_source_to_creature(traverse->current,
		  find_light_source_radius(traverse->current));
    }

void add_light_source_to_spot(struct level *l, long x, long y, long radius)
    {
    struct light_source_struct *light_source;
    light_source = (struct light_source_struct *) calloc(1, sizeof(struct
      light_source_struct));
#ifdef DEBUG
    light_source->me = light_source;
#endif
    light_source->x = x;
    light_source->y = y;
    light_source->holder = NULL;
    light_source->radius = radius;
    light_source->is_on = TRUE;
    light_source->next = l->light_sources;
    l->light_sources = light_source;
    }

/* This may have bugs (potentially freeing something it uses.  */

void one_use_light_sources()
    {
    struct light_source_struct *traverse;
    struct small_item *light;
    if (current_level->light_sources != NULL)
	for(traverse = current_level->light_sources; traverse != NULL; traverse
	  = traverse->next)
	    if (traverse->holder != NULL)
		{
		light = find_light_source(traverse->holder);
		if (light != NULL)
		    {
		    if (flag_check(light->flags, DISAPPEAR_LAST_USE))
			{
			if (traverse->holder->extended != NULL)
			    if (traverse->holder->extended->wear_wield != NULL)
				{
				find_and_remove_id(
				  &traverse->holder->extended->wear_wield,
				  light->identity);
				if (find_light_source(traverse->holder) == 0)
				while (remove_creatures_light_source(
				  current_level, traverse->holder));
				}
			}
		    else
			{
			--light->uses;
			if (light->uses == 0)
			    {
			    if (traverse->holder == player)
				pmsg("Your light goes out.");
			    flag_reset(&light->flags, ACTIVATED);
			    while (remove_creatures_light_source(current_level,
			      traverse->holder));
			    }
			}
		    }
		}
    }

void status_message(struct small_creature *c, long status)
    {
    if ((0 <= status) && (status < NUM_STATI) && (c != NULL))
	{
	if (status_check(c, status) > 0)
	    switch(status)
		{
		case STATUS_HUNGRY:
		    switch(status_check(c, status))
			{
			case 1:
			    pmsg("You are feeling hungry.");
			break;
			case 2:
			    pmsg("You feel very hungry.");
			break;
			case 3:
			    pmsg("You are ravenous!");
			break;
			case 4:
			    pmsg("You are weak!");
			break;
			case 5:
			    pmsg("You're starving!");
			break;
			}
		break;
		case STATUS_THIRSTY:
		    switch(status_check(c, status))
			{
			case 1:
			    pmsg("You feel thirsty.");
			break;
			case 2:
			    pmsg("You feel very thirsty.");
			break;
			case 3:
			    pmsg("You are dehydrated.");
			break;
			case 4:
			    pmsg("You are severely dehydrated.");
			break;
			case 5:
			    pmsg("You are dying of thirst!");
			break;
			}
		break;
		case STATUS_TIRED:
		    switch(status_check(c, status))
			{
			case 1:
			    pmsg("You feel a bit drowsy.");
			break;
			case 2:
			    pmsg("You feel tired.");
			break;
			case 3:
			    pmsg("You feel very sleepy.");
			break;
			case 4:
			    pmsg("You are barely awake!");
			break;
			case 5:
			    pmsg("You are dying for lack of sleep!");
			break;
			}
		break;
		case STATUS_FATIGUED:
		    switch(status_check(c, status))
			{
			case 1:
			    pmsg("You are fatigued.");
			break;
			case 2:
			    pmsg("You are worn out.");
			break;
			case 3:
			    pmsg("You are exhausted!");
			break;
			case 4:
			    pmsg("You are totally exhausted!");
			break;
			case 5:
			    pmsg("You are dangerously fatigued!");
			break;
			}
		break;
		case STATUS_COLD:
		    switch(status_check(c, status))
			{
			case 1:
			    pmsg("You are a bit chilly.");
			break;
			case 2:
			    pmsg("You are cold.");
			break;
			case 3:
			    pmsg("You are chilled to the bone!");
			break;
			case 4:
			    pmsg("You have come down with hypothermia!");
			break;
			case 5:
			    pmsg("You are dying of hypothermia!");
			break;
			}
		break;
		}
	}
    }

void add_item_to_level(struct level *l, long item_id)
    {
    long x, y;
    struct ilist *this_ilist;
    struct small_item *this_item;
    x = trandom() % MAX_X;
    y = trandom() % MAX_Y;
    while ((!(flag_check(l->space[x][y].flags, CAN_OCCUPY))) ||
      (l->space[x][y].items != NULL))
	{
	x = trandom() % MAX_X;
	y = trandom() % MAX_Y;
	}
    this_item = generate_item(l, item_id);
    if (flag_check(this_item->flags, POOR_QUALITY))
	small_item_make_normal(this_item);
    this_ilist = (struct ilist *) calloc(1, sizeof(struct ilist));
#ifdef DEBUG
    this_ilist->me = this_ilist;
#endif
    this_ilist->next = l->space[x][y].items;
    this_ilist->current = this_item;
    l->space[x][y].items = this_ilist;
    }

long is_on_clist(struct small_creature *c, struct clist *cl)
    {
    struct clist *traverse;
    if (cl != NULL)
	for(traverse = cl; traverse != NULL; traverse = traverse->next)
	    if (traverse->current == c)
		return TRUE;
    return FALSE;
    }

void find_name(struct small_creature *c, char *namefile)
    {
    char real_filename[256], this_line[256];
    long my_continue, num_excerpts, chosen_excerpt, i, success;
    FILE *infile;
    success = TRUE;
    strcpy(real_filename, LIB_DIR);
    strcat(real_filename, namefile);
    if (c != NULL)
	{
	if (!(infile = fopen(real_filename, "r")))
	    {
	    perror("Attempt failed to open file:");
	    perror(real_filename);
	    success = FALSE;
	    }
	else
	    {
	    my_continue = 1;
	    num_excerpts = 0;
	    while (my_continue)
		{
		my_continue = ((my_continue) && (fgets(this_line, 256, infile)));
		if (strcnt(this_line, "##"))
		    ++num_excerpts;
		}
	    fclose(infile);
	    if (num_excerpts == 0)
		{
		perror("Error: excerpt requested from file with no excerpts");
		success = FALSE;
		}
	    else
		{
		chosen_excerpt = (trandom() % num_excerpts) + 1;
		my_continue = TRUE;
		infile = fopen(real_filename, "r");
		strcpy(this_line, "");
		for(i = 0; i < chosen_excerpt; fgets(this_line, 256, infile))
		    if (strcnt(this_line, "##"))
			++i;
		if (c->name != NULL)
		    free_text_struct(c->name);
		c->name = (struct text_struct *) calloc(1, sizeof(struct
		  text_struct));
#ifdef DEBUG
		c->name->me = c->name;
#endif
		c->name->next = NULL;
		for(i = 0; i < strlen(this_line); ++i)
		    if ((this_line[i] == 10) || (this_line[i] == 13))
			this_line[i] = 0;
		strcpy(c->name->text, this_line);
		strcat(c->name->text, " (");
		strcat(c->name->text, master_creatures[c->identity].species);
		strcat(c->name->text, ")");
		fclose(infile);
		}
	    }
	}
    }

long can_fly(struct small_creature *c)
    {
    if (c != NULL)
	{
	if (find_vehicle(c) != NULL)
	    if (find_vehicle(c)->identity == HOT_AIR_BALLOON)
		return TRUE;
	if (c->extended != NULL)
	    if (c->extended->mount != NULL)
		if (flag_check(c->extended->mount->flags, FLYING))
		    return TRUE;
	return ((flag_check(c->flags, FLYING)) || (status_check(c,
	  STATUS_FLYING) > 0));
	}
    else
	return 0;
    }

struct small_item *find_vehicle(struct small_creature *c)
    {
    struct ilist *traverse;
    struct master_item real_vehicle;
    if (c != NULL)
	if (c->extended != NULL)
	    if (c->extended->wear_wield != NULL)
		for(traverse = c->extended->wear_wield; traverse != NULL;
		  traverse = traverse->next)
		    {
		    real_vehicle = orig_items[traverse->current->identity];
		    if (flag_check(real_vehicle.slots, VEHICLE))
			return traverse->current;
		    }
    return NULL;
    }

long language_check(struct small_creature *c, long language)
    {
    long skill;
    skill = 0;
    if ((language >= JANRA) && (language <= YEDIDIA))
	{
	if (c != NULL)
	    {
	    if (c->extended != NULL)
		{
		switch(language)
		    {
		    case JANRA:
			skill = LANGUAGE_JANRA;
		    break;
		    case JEC:
			skill = LANGUAGE_JEC;
		    break;
		    case SHAL:
			skill = LANGUAGE_SHAL;
		    break;
		    case TUZ:
			skill = LANGUAGE_TUZ;
		    break;
		    case URVANOVESTILLI:
			skill = LANGUAGE_URVANOVESTILLI;
		    break;
		    case YEDIDIA:
			skill = LANGUAGE_YEDIDIA;
		    break;
		    }
		return (skill_check(c, c->extended->skills->av[skill],
		  DIFFICULT, SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE) > 0);
		}
	    else
		return 0;
	    }
	else
	    return 0;
	}
    else
	return 0;
    }

void append_text_to_text_struct(struct text_struct **text, char *addition)
    {
    struct text_struct *traverse;
    if (*text == NULL)
	{
	*text = (struct text_struct *) calloc(1, sizeof(struct text_struct));
#ifdef DEBUG
	(*text)->me = *text;
#endif
	(*text)->next = NULL;
	strcpy((*text)->text, addition);
	}
    else if ((*text)->next == NULL)
	{
	(*text)->next = (struct text_struct *) calloc(1, sizeof(struct
	  text_struct));
#ifdef DEBUG
	(*text)->next->me = (*text)->next;
#endif
	(*text)->next->next = NULL;
	strcpy((*text)->next->text, addition);
	}
    else
	{
	for(traverse = *text; traverse->next != NULL; traverse =
	  traverse->next);
	traverse->next = (struct text_struct *) calloc(1, sizeof(struct
	  text_struct));
#ifdef DEBUG
	traverse->next->me = traverse->next;
#endif
	traverse->next->next = NULL;
	strcpy(traverse->next->text, addition);
	}
    }

void encounter_internal(struct level *l, long encounter_type, long success)
    {
    long i, x, y, my_continue;
    struct small_creature *animal;
    for(i = 0; i < success; ++i)
	{
	animal = make_random_creature(encounter_type);
	my_continue = TRUE;
	while (my_continue)
	    {
	    x = trandom() % MAX_X;
	    y = trandom() % MAX_Y;
	    my_continue =
	      ((flag_check(l->space[x][y].flags, OCCUPIED)) ||
	      (!(flag_check(l->space[x][y].flags,
	      CAN_OCCUPY))) || ((flag_check(animal->flags, WATER_ONLY)) &&
	      (!(flag_check(l->space[x][y].flags, IS_WATER)))));
	    }
	no_hostile_add_creature(l, animal, x, y);
	check_addition(l, animal->identity, x, y);
	if ((0 <= encounter_type) && (encounter_type < NUM_CREATURES))
	    if (flag_check(master_creatures[i].flags, UNIQUE))
		i = success;
	}
    hostility_check(l, player);
    player->nextaction = game_variables->elapsed_units;
    }

void human_encounter_internal(struct level *l, long encounter_type, long
  success)
    {
    long i, x, y, my_continue;
    struct small_creature *person;
    for(i = 0; i < success; ++i)
	{
	if ((trandom() % 2 == 0) || (encounter_type == NORKRIN))
	    person = new_npc_generic(encounter_type);
	else
	    person = new_adventurer();
	my_continue = TRUE;
	while (my_continue)
	    {
	    x = trandom() % MAX_X;
	    y = trandom() % MAX_Y;
	    my_continue =
	      ((flag_check(l->space[x][y].flags, OCCUPIED)) ||
	      (!(flag_check(l->space[x][y].flags,
	      CAN_OCCUPY))));
	    }
	no_hostile_add_creature(l, person, x, y);
	while (trandom() % 4 != 0)
	    {
	    if ((trandom() % 2 == 0) || (encounter_type == NORKRIN))
		person = new_npc_generic(encounter_type);
	    else
		person = new_adventurer();
	    while ((flag_check(l->space[x][y].flags, OCCUPIED)) ||
	      (!(flag_check(l->space[x][y].flags,
	      CAN_OCCUPY))))
		{
		switch(trandom() % 4)
		    {
		    case 0:
			++x;
		    break;
		    case 1:
			--x;
		    break;
		    case 2:
			++y;
		    break;
		    case 3:
			--y;
		    break;
		    }
		x = min(max(x, 0), MAX_X - 1);
		y = min(max(y, 0), MAX_Y - 1);
		}
	no_hostile_add_creature(l, person, x, y);
	    }
	if ((0 <= encounter_type) && (encounter_type < NUM_CREATURES))
	    if (flag_check(master_creatures[i].flags, UNIQUE))
		i = success;
	}
    hostility_check(l, player);
    player->nextaction = game_variables->elapsed_units;
    }

long is_tired(struct small_creature *c)
    {
    return (status_check(c, STATUS_TIRED) > 0);
    }

void save_player()
    {
    int player_file;
    clear();
    pmsg("Allocate experience to skills before saving?");
    while (confirm_silent())
	{
	allocate_experience(player);
	pmsg("Allocate experience in another skill?");
	}
    player_file = open(".tmsrc", O_WRONLY | O_CREAT | O_TRUNC, 0600);
    if (player_file == -1)
	{
	printw("I couldn't save your character.  But I can let you play this game...\n");
	fs_anykey();
	}
    else
	{
	write(player_file, player, sizeof(*player));
	write(player_file, player->extended, sizeof(*player->extended));
	write(player_file, player->extended->faith,
	  sizeof(*player->extended->faith));
	if (player->extended->faith->breath_prayer != NULL)
	    write(player_file, player->extended->faith->breath_prayer,
	      sizeof(*player->extended->faith->breath_prayer));
	write(player_file, player->extended->stati,
	  sizeof(*player->extended->stati));
	write(player_file, player->extended->skills,
	  sizeof(*player->extended->skills));
	close(player_file);
	}
    }

void load_player()
    {
    int player_file;
    struct clist *mcl;
    player_file = open(".tmsrc", O_RDONLY, 0600);
    if (player_file == -1)
	{
	fprintf(stderr, "Character not found!  Exiting...\n");
	my_finish(1);
	}
    else
	{
	player = (struct small_creature *) calloc(1, sizeof(struct
	  small_creature));
	read(player_file, player, sizeof(*player));
#ifdef DEBUG
	player->me = player;
#endif
	player->extended = (struct master_creature *) calloc(1, sizeof(struct
	  master_creature));
	read(player_file, player->extended, sizeof(*player->extended));
	player->extended->faith = (struct faith_struct *) calloc(1,
	  sizeof(struct faith_struct));
	read(player_file, player->extended->faith,
	  sizeof(*player->extended->faith));
#ifdef DEBUG
	player->extended->faith->me = player->extended->faith;
#endif
	if (player->extended->faith->breath_prayer != NULL)
	    {
	    player->extended->faith->breath_prayer = (struct text_struct *)
	      calloc(1, sizeof(struct text_struct));
	    read(player_file, player->extended->faith->breath_prayer,
	      sizeof(*player->extended->faith->breath_prayer));
#ifdef DEBUG
	    player->extended->faith->breath_prayer->me =
	      player->extended->faith->breath_prayer;
#endif
	    }
	player->extended->stati = (struct adj_struct *) calloc(1, sizeof(struct
	  adj_struct));
	read(player_file, player->extended->stati,
	  sizeof(*player->extended->stati));
#ifdef DEBUG
	player->extended->stati->me = player->extended->stati;
#endif
	player->extended->skills = (struct skills_struct *) calloc(1, sizeof(
	  struct skills_struct));
	read(player_file, player->extended->skills,
	  sizeof(*player->extended->skills));
#ifdef DEBUG
	player->extended->skills->me = player->extended->skills;
#endif
	mcl = (struct clist *) calloc(1, sizeof(struct clist));
#ifdef DEBUG
	mcl->me = mcl;
#endif
	mcl->next = master_clist;
	mcl->current = player;
	master_clist = mcl;
	close(player_file);
	}
    }

long one_move(struct small_creature *c, long *result_x, long *result_y)
    {
    long curr_x, curr_y, my_continue;
    char command;
    pmsg("Which direction?");
    curr_x = 0;
    curr_y = 0;
    my_continue = TRUE;
    for(command=0;(my_continue);)
	{
	my_continue = FALSE;
	switch(command)
	    {
	    case LEFT: case SECOND_LEFT: --curr_x; break;
	    case DOWN: case SECOND_DOWN: ++curr_y; break;
	    case UP: case SECOND_UP: --curr_y; break;
	    case RIGHT: case SECOND_RIGHT: ++curr_x; break;
	    case UP_LEFT: case SECOND_UP_LEFT: --curr_x; --curr_y; break;
	    case UP_RIGHT: case SECOND_UP_RIGHT: ++curr_x; --curr_y; break;
	    case DOWN_LEFT: case SECOND_DOWN_LEFT: --curr_x; ++curr_y; break;
	    case DOWN_RIGHT: case SECOND_DOWN_RIGHT: ++curr_x; ++curr_y; break;
	    case '?':
		pmsg("Use the motion commands to select a direction.");
		my_continue = TRUE;
	    break;
	    case ESCAPE: pmsg("Aborted."); break;
	    default:
		my_continue = TRUE;
	    break;
	    }
	if (my_continue)
	    command = tgetc();
	}
    *result_x = curr_x;
    *result_y = curr_y;
    return (command != ESCAPE);
    }

struct small_creature *one_move_loc_creature(struct small_creature *c)
    {
    long x, y;
    if (one_move(c, &x, &y))
	{
	if ((0 <= c->x + x) && (c->x + x < MAX_X) && (0 <= c->y + y) && (c->y
	  + y <= MAX_Y))
	    {
	    if (current_level->space[c->x + x][c->y + y].creatures != NULL)
		return current_level->space[c->x + x][c->y +
		  y].creatures->current;
	    else
		return NULL;
	    }
	else
	    return NULL;
	}
    else
	return NULL;
    }

void item_desc(struct small_item *i, char *description)
    {
    if (i != NULL)
	{
	if ((!(i->tname == NULL)) &&
	  ((!(flag_check(i->flags2,
	  DONT_PRINT_NAME_UNLESS_IDENTIFIED))) ||
	  (flag_check(i->flags, IDENTIFIED))))
	    strcpy(description, i->tname->text);
	else if (!(i->name == ST_DEFAULT))
	    find_string(i->name, description);
	else if (flag_check(i->flags, IDENTIFIED))
	    strcpy(description,
	      orig_items[i->identity].identified);
	else
	    strcpy(description,
	      orig_items[i->identity].nonidentified);
	}
    else
	strcpy(description, "");
    }

void describe_item(struct small_item *i)
    {
    char item_name[256], message[256], value_string[256];
    if (i != NULL)
	{
	clear();
	item_desc(i, item_name);
	strcat(item_name, ":");
	fs_pmsg(item_name);
	fs_pmsg("");
	if (flag_check(i->flags, IDENTIFIED))
	    {
	    if (flag_check(i->flags, FINE_QUALITY))
		fs_pmsg("Is of fine quality.");
	    else if (flag_check(i->flags, POOR_QUALITY))
		fs_pmsg("Is of poor quality.");
	    }
	if ((flag_check(i->flags, CONTAINER)) && (((flag_check(i->flags,
	  IDENTIFIED)) && (flag_check(i->flags, MAY_HAVE_COMPARTMENT))) ||
	  (!(flag_check(i->flags, MAY_HAVE_COMPARTMENT)))))
	    {
	    if (flag_check(i->flags, MAY_HAVE_COMPARTMENT))
		fs_pmsg("Has secret compartment.");
	    else
		fs_pmsg("Is a container.");
	    }
	if (flag_check(i->flags, IS_OPEN))
	    fs_pmsg("Is open.");
	if (flag_check(i->flags, LOCKED))
	    fs_pmsg("Is locked.");
	if (flag_check(i->flags, ACTIVATED))
	    fs_pmsg("Is activated.");
	if (i->writing != NULL)
	    fs_pmsg("Has writing.");
	if (flag_check(i->flags2, IS_WEAPON))
	    if (i->other > 0)
		fs_pmsg("Is poisoned.");
	if (flag_check(i->flags2, APPRAISED))
	    {
	    money_string(i->value, value_string);
	    sprintf(message, "Is of value %s", value_string);
	    fs_pmsg(message);
	    }
	if (i->uses > 0)
	    {
	    sprintf(message, "Has %ld uses remaining.", i->uses);
	    fs_pmsg(message);
	    }
	fs_pmsg("");
	if (flag_check(i->flags, IDENTIFIED))
	    printw(orig_items[i->identity].iddesc);
	else
	    printw(orig_items[i->identity].noniddesc);
	fs_anykey();
	clear();
	}
    }

void pass_out(struct small_creature *c, long amount)
    {
    char creature_title[256], message[256];
    if (status_check(c, STATUS_PASSED_OUT) <= 0)
	{
	if (c == player)
	    pmsg("You pass out!");
	else if (can_see(player, c->x, c->y))
	    {
	    find_title(c, creature_title);
	    sprintf(message, "%s passes out!", creature_title);
	    pmsg(message);
	    }
	}
    status_add(c, STATUS_PASSED_OUT, amount);
    flag_set(&c->flags, CANT_SEE);
    }

void come_to(struct small_creature *c)
    {
    char creature_title[256], message[256];
    if (c == player)
	pmsg("You come to.");
    else if (can_see(player, c->x, c->y))
	{
	find_title(c, creature_title);
	sprintf(message, "%s comes to.", creature_title);
	pmsg(message);
	}
    status_add(c, STATUS_PASSED_OUT, -100);
    flag_reset(&c->flags, CANT_SEE);
    }

struct small_creature *appoint_successor(struct small_creature *c)
    {
    struct clist *traverse;
    long found_successor, my_continue;
    char candidate_title[256], message[256];
    if (c->followed_by != NULL)
	{
	found_successor = FALSE;
	for(traverse = c->followed_by; traverse != NULL; traverse =
	  traverse->next)
	    if (flag_check(traverse->current->flags, HUMAN))
		found_successor = TRUE;
	if (found_successor)
	    {
	    pmsg("Appoint a successor?");
	    if (confirm_silent())
		{
		my_continue = TRUE;
		while (my_continue)
		    {
		    for(traverse = c->followed_by; traverse != NULL; traverse =
		      traverse->next)
			if (flag_check(traverse->current->flags, HUMAN))
			    {
			    find_title(traverse->current, candidate_title);
			    sprintf(message,
"Candidate: %s: appoint as successor?", candidate_title);
			    pmsg(message);
			    if (confirm_silent())
				return traverse->current;
			    }
		    pmsg("Those were all of your human followers.  Go through the list again?");
		    my_continue = confirm_silent();
		    }
		return FALSE;
		}
	    }
	else
	    return NULL;
	}
    else
	return NULL;
    return NULL;
    }

long has_hummer(struct small_creature *c)
    {
    struct ilist *traverse;
    if (c != NULL)
	if (c->extended != NULL)
	    if (c->extended->wear_wield != NULL)
		for(traverse = c->extended->wear_wield; traverse != NULL;
		  traverse = traverse->next)
		    if (traverse->current->identity == HUMMER)
			return TRUE;
    return FALSE;    
    }

void describe_creature(struct small_creature *c)
    {
    long result, age_class, i, wounded, temp;
    char message[256], pronoun[256], mount_title[256];
    if (c != NULL)
	{
	if (flag_check(c->flags, HIDDEN))
	    result = skill_check(player, player->extended->skills->av[SEARCH],
	      DIFFICULT, SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE);
	else
	    result = skill_check(player, player->extended->skills->
	      av[KEEN_EYESIGHT], EASY,
	      SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE);
	if ((result > 0) && (c != player))
	    {
	    if (result == 1)
		fs_pmsg("Some creature or person -- you can't tell what.");
	    else if (result == 2)
		{
		find_title(c, message);
		strcat(message, ":");
		cap_fs_pmsg(message);
		}
	    else
		{
		find_title(c, message);
		strcat(message, ".");
		cap_fs_pmsg(message);
		find_pronoun(c, pronoun);
		if (master_creatures[c->identity].gender != NEUTER)
		    {
		    if (flag_check(c->flags, C_FEMALE))
			sprintf(message, "%s is female.", pronoun);
		    else
			sprintf(message, "%s is male.", pronoun);
		    cap_fs_pmsg(message);
		    }
		age_class = MIDDLE_AGED;
		if (c->extended != NULL)
		    age_class = c->extended->age_class;
		else
		    for(i = 0; i < NUM_AGE_CLASSES; ++i)
			if (c->age > age_class_years[FIRST_NONHUMAN][i])
			    age_class = i + 1;
		switch(age_class)
		    {
		    case INFANT:
			sprintf(message, "%s is an infant.", pronoun);
		    break;
		    case CHILD:
			sprintf(message, "%s is a child.", pronoun);
		    break;
		    case YOUNG_ADULT:
			sprintf(message, "%s is a young adult.", pronoun);
		    break;
		    case MIDDLE_AGED:
			sprintf(message, "%s is middle aged.", pronoun);
		    break;
		    case OLD:
			sprintf(message, "%s is old.", pronoun);
		    break;
		    case ANCIENT:
			sprintf(message, "%s is ancient.", pronoun);
		    break;
		    default:
		        sprintf(message, "Inappropriate age class.");
		    break;
		    }
		cap_fs_pmsg(message);
		if (c->extended != NULL)
		    if (c->extended->mount != NULL)
			{
			find_title(c->extended->mount, mount_title);
			sprintf(message, "%s is riding %s.", pronoun,
			  mount_title);
			}
		if (c->extended != NULL)
		    wounded = status_check(c, STATUS_WOUNDED);
		else
		    wounded = c->wounded;
		if ((wounded >= 0) && (skill_check(player,
		  player->extended->skills->av[ANATOMY], DIFFICULT,
		  SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE) > 0))
		    {
		    if (wounded == 0)
			sprintf(message, "%s is healthy.", pronoun);
		    else if (wounded <= 1)
			sprintf(message, "%s is slightly bruised.", pronoun);
		    else if (wounded <= 3)
			sprintf(message, "%s has mild wounds.", pronoun);
		    else if (wounded <= 7)
			sprintf(message, "%s is moderately wounded.", pronoun);
		    else if (wounded <= 15)
			sprintf(message, "%s is heavily wounded.", pronoun);
		    else if (wounded <= 31)
			sprintf(message, "%s is severely wounded.", pronoun);
		    else if (wounded <= 63)
			sprintf(message, "%s is grievously wounded.", pronoun);
		    else
			sprintf(message, "%s is barely alive.", pronoun);
		    cap_fs_pmsg(message);
		    }
		if (c->hostile != NULL)
		    {
		    sprintf(message, "%s appears agitated.", pronoun);
		    cap_fs_pmsg(message);
		    }
		if (c->extended != NULL)
		    {
		    if ((c->extended->pack != NULL) || (c->extended->wear_wield
		      != NULL))
			{
			temp = 'a';
			if (c->extended->wear_wield != NULL)
			    {
			    fs_pmsg("Equipment in use:");
			    print_ilist_internal(c->extended->wear_wield, 4,
			      &temp, FULLSCREEN);
			    }
			temp = 'a';
			if (c->extended->pack != NULL)
			    {
			    fs_pmsg("Possessions carried:");
			    print_ilist_internal(c->extended->pack, 4,
			      &temp, FULLSCREEN);
			    }
			}
		    else
			{
			sprintf(message, "%s is empty-handed.", pronoun);
			cap_fs_pmsg(message);
			}
		    }
		else
		    {
		    sprintf(message, "%s is empty-handed.", pronoun);
		    cap_fs_pmsg(message);
		    }
		}
	    fs_pmsg("");
	    }
	else if (c == player)
	    fs_pmsg("You");
	}
    }
